@@ -1,5 +1,5 @@
 /*************************************************************************** 
- * RT2400 SourceForge Project - http://rt2400.sourceforge.net              * 
+ * RT2400/RT2500 SourceForge Project - http://rt2x00.serialmonkey.com      * 
  *                                                                         * 
  *   This program is free software; you can redistribute it and/or modify  * 
  *   it under the terms of the GNU General Public License as published by  * 
@@ -17,24 +17,67 @@
  *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             * 
  *                                                                         * 
  *   Licensed under the GNU GPL                                            * 
- *   Original code supplied under license from RaLink Inc, 2003.           * 
+ *   Original code supplied under license from RaLink Inc, 2004.           * 
  ***************************************************************************/ 
 
  /*************************************************************************** 
  *      Module Name: assoc.c 
  *              
- *      Abstract: Association State Machine 
+ *      Abstract: 
  *              
  *      Revision History: 
  *      Who             When            What 
  *      --------        -----------     ----------------------------- 
- *      MarkW           9th  Feb 04     Baseline of code
- *      defekt          13th Feb 04     ATOMIC kmalloc fix
+ *      MarkW           8th  Dec 04     Baseline code  
+ * 		MarkW			5th  Jun 05		Fix no-SSID broadcasting assoc.
  ***************************************************************************/ 
 
-
 #include "rt_config.h"
 
+UCHAR	CipherSuiteWpaTkip[] = {
+		0x00, 0x50, 0xf2, 0x01,	// oui
+		0x01, 0x00,				// Version
+		0x00, 0x50, 0xf2, 0x02,	// Multicast
+		0x01, 0x00,				// Number of unicast
+		0x00, 0x50, 0xf2, 0x02,	// unicast
+		0x01, 0x00,				// number of authentication method
+		0x00, 0x50, 0xf2, 0x01	// authentication
+		};
+UCHAR	CipherSuiteWpaTkipLen = (sizeof(CipherSuiteWpaTkip) / sizeof(UCHAR));
+
+UCHAR	CipherSuiteWpaAes[] = {
+		0x00, 0x50, 0xf2, 0x01,	// oui
+		0x01, 0x00,				// Version
+		0x00, 0x50, 0xf2, 0x04,	// Multicast
+		0x01, 0x00,				// Number of unicast
+		0x00, 0x50, 0xf2, 0x04,	// unicast
+		0x01, 0x00,				// number of authentication method
+		0x00, 0x50, 0xf2, 0x01	// authentication
+		};
+UCHAR	CipherSuiteWpaAesLen = (sizeof(CipherSuiteWpaAes) / sizeof(UCHAR));
+
+UCHAR	CipherSuiteWpaPskTkip[] = {
+		0x00, 0x50, 0xf2, 0x01,	// oui
+		0x01, 0x00,				// Version
+		0x00, 0x50, 0xf2, 0x02,	// Multicast
+		0x01, 0x00,				// Number of unicast
+		0x00, 0x50, 0xf2, 0x02,	// unicast
+		0x01, 0x00,				// number of authentication method
+		0x00, 0x50, 0xf2, 0x02	// authentication
+		};
+UCHAR	CipherSuiteWpaPskTkipLen = (sizeof(CipherSuiteWpaPskTkip) / sizeof(UCHAR));
+
+UCHAR	CipherSuiteWpaPskAes[] = {
+		0x00, 0x50, 0xf2, 0x01,	// oui
+		0x01, 0x00,				// Version
+		0x00, 0x50, 0xf2, 0x04,	// Multicast
+		0x01, 0x00,				// Number of unicast
+		0x00, 0x50, 0xf2, 0x04,	// unicast
+		0x01, 0x00,				// number of authentication method
+		0x00, 0x50, 0xf2, 0x02	// authentication
+		};
+UCHAR	CipherSuiteWpaPskAesLen = (sizeof(CipherSuiteWpaPskAes) / sizeof(UCHAR));
+
 /*  
     ==========================================================================
     Description: 
@@ -44,7 +87,7 @@
     Note:
         The state machine looks like the following 
         
-    ASSOC_IDLE               ASSOC_WAIT_RSP           REASSOC_WAIT_RSP           DISASSOC_WAIT_RSP
+                               ASSOC_IDLE               ASSOC_WAIT_RSP             REASSOC_WAIT_RSP             DISASSOC_WAIT_RSP
     MT2_MLME_ASSOC_REQ       mlme_assoc_req_action    invalid_state_when_assoc   invalid_state_when_assoc       invalid_state_when_assoc
     MT2_MLME_REASSOC_REQ     mlme_reassoc_req_action  invalid_state_when_reassoc invalid_state_when_reassoc     invalid_state_when_reassoc
     MT2_MLME_DISASSOC_REQ    mlme_disassoc_req_action mlme_disassoc_req_action   mlme_disassoc_req_action       mlme_disassoc_req_action
@@ -100,17 +143,9 @@ VOID AssocStateMachineInit(
     StateMachineSetAction(S, DISASSOC_WAIT_RSP, MT2_DISASSOC_TIMEOUT, (STATE_MACHINE_FUNC)DisassocTimeoutAction);
 
     // initialize the timer
-    init_timer(&pAd->Mlme.AssocAux.AssocTimer);
-	pAd->Mlme.AssocAux.AssocTimer.data = (unsigned long)pAd;
-	pAd->Mlme.AssocAux.AssocTimer.function = &AssocTimeout;				/* timer handler */
-
-	init_timer(&pAd->Mlme.AssocAux.ReassocTimer);
-	pAd->Mlme.AssocAux.ReassocTimer.data = (unsigned long)pAd;
-	pAd->Mlme.AssocAux.ReassocTimer.function = &ReassocTimeout;				/* timer handler */
-
-	init_timer(&pAd->Mlme.AssocAux.DisassocTimer);
-	pAd->Mlme.AssocAux.DisassocTimer.data = (unsigned long)pAd;
-	pAd->Mlme.AssocAux.DisassocTimer.function = &DisassocTimeout;				/* timer handler */
+    RTMPInitTimer(pAd, &pAd->Mlme.AssocAux.AssocTimer, AssocTimeout);
+    RTMPInitTimer(pAd, &pAd->Mlme.AssocAux.ReassocTimer, ReassocTimeout);
+    RTMPInitTimer(pAd, &pAd->Mlme.AssocAux.DisassocTimer, DisassocTimeout);
 }
 
 /*
@@ -123,7 +158,7 @@ VOID AssocStateMachineInit(
     ==========================================================================
  */
 VOID AssocTimeout(
-	IN unsigned long data) 
+    IN	unsigned long data) 
 {
     RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;
     DBGPRINT(RT_DEBUG_TRACE,"ASSOC - enqueue MT2_ASSOC_TIMEOUT \n");
@@ -141,7 +176,7 @@ VOID AssocTimeout(
     ==========================================================================
  */
 VOID ReassocTimeout(
-	IN unsigned long data) 
+    IN	unsigned long data) 
 {
     RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;
     DBGPRINT(RT_DEBUG_TRACE,"ASSOC - enqueue MT2_REASSOC_TIMEOUT \n");
@@ -159,7 +194,7 @@ VOID ReassocTimeout(
     ==========================================================================
  */
 VOID DisassocTimeout(
-	IN unsigned long data) 
+    IN	unsigned long data) 
 {
     RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)data;
     DBGPRINT(RT_DEBUG_TRACE,"ASSOC - enqueue MT2_DISASSOC_TIMEOUT \n");
@@ -192,23 +227,35 @@ VOID MlmeAssocReqAction(
 {
     MACADDR          ApAddr;
     MACHDR           AssocHdr;
-    UCHAR            SsidIe = IE_SSID, RateIe = IE_SUPP_RATES;
+    UCHAR            SsidIe = IE_SSID, RateIe = IE_SUPP_RATES,WpaIe = IE_WPA, ExtRateIe = IE_EXT_SUPP_RATES;
     USHORT           ListenIntv;
     ULONG            Timeout;
     USHORT           CapabilityInfo;
     UCHAR           *OutBuffer = NULL;
+    NDIS_STATUS      NStatus;
     ULONG            FrameLen = 0;
+	ULONG			 tmp;
+	UCHAR			 VarIesOffset;
 
+	// Block all authentication request durning WPA block period
+	if (pAd->PortCfg.bBlockAssoc == TRUE)
+	{
+        DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Block Assoc request durning WPA block period!\n");
+        pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
+        MlmeCntlConfirm(pAd, MT2_ASSOC_CONF, MLME_STATE_MACHINE_REJECT);
+	}	
     // check sanity first
-    if (MlmeAssocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &ApAddr, &CapabilityInfo, &Timeout, &ListenIntv)) 
+    else if (MlmeAssocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &ApAddr, &CapabilityInfo, &Timeout, &ListenIntv)) 
     {
-        del_timer_sync(&pAd->Mlme.AssocAux.AssocTimer);
+        RTMPCancelTimer(&pAd->Mlme.AssocAux.AssocTimer);
         COPY_MAC_ADDR(&pAd->Mlme.AssocAux.Addr, &ApAddr);
+		// Mask out unnecessary capability information
+		CapabilityInfo &= SUPPORTED_CAPABILITY_INFO; // pAd->PortCfg.SupportedCapabilityInfo;
         pAd->Mlme.AssocAux.CapabilityInfo = CapabilityInfo;
         pAd->Mlme.AssocAux.ListenIntv = ListenIntv;
 
-        OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_ATOMIC);
-        if(OutBuffer == NULL)
+        NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
+        if (NStatus != NDIS_STATUS_SUCCESS) 
         {
             DBGPRINT(RT_DEBUG_TRACE,"ASSOC - MlmeAssocReqAction() allocate memory failed \n");
             pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
@@ -216,23 +263,177 @@ VOID MlmeAssocReqAction(
             return;
         }
         
+		// Add by James 03/06/27
+		pAd->PortCfg.AssocInfo.Length = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION); //+ sizeof(NDIS_802_11_FIXED_IEs); 	// Filled in assoc request
+		pAd->PortCfg.AssocInfo.AvailableRequestFixedIEs =
+			NDIS_802_11_AI_REQFI_CAPABILITIES | NDIS_802_11_AI_REQFI_LISTENINTERVAL | NDIS_802_11_AI_REQFI_CURRENTAPADDRESS;
+		pAd->PortCfg.AssocInfo.RequestFixedIEs.Capabilities = CapabilityInfo;
+		pAd->PortCfg.AssocInfo.RequestFixedIEs.ListenInterval = ListenIntv;		
+		memcpy(pAd->PortCfg.AssocInfo.RequestFixedIEs.CurrentAPAddress, &AssocHdr, sizeof(NDIS_802_11_MAC_ADDRESS));
+		pAd->PortCfg.AssocInfo.OffsetRequestIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION);		// No request Variables IEs
+		
+		// First add SSID
+		VarIesOffset = 0;
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &SsidIe, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &pAd->PortCfg.SsidLen, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, pAd->PortCfg.Ssid, pAd->PortCfg.SsidLen);
+		VarIesOffset += pAd->PortCfg.SsidLen;
+
+		// Second add Supported rates
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &RateIe, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &pAd->PortCfg.SupportedRatesLen, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, pAd->PortCfg.SupportedRates, pAd->PortCfg.SupportedRatesLen);
+		VarIesOffset += pAd->PortCfg.SupportedRatesLen;
+		// End Add by James
+
         DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Send ASSOC request...\n");
         MgtMacHeaderInit(pAd, &AssocHdr, SUBTYPE_ASSOC_REQ, 0, &ApAddr, &ApAddr);
-        MakeOutgoingFrame(OutBuffer, &FrameLen,
-                          sizeof(MACHDR), (CHAR *)&AssocHdr,
-                          2, &CapabilityInfo,                   // capability
-                          2, &ListenIntv,                       // listen interval
-                          1, &SsidIe,                           // ssid
-                          1, &pAd->PortCfg.SsidLen, 
-                          pAd->PortCfg.SsidLen, pAd->PortCfg.Ssid,
-                          1, &RateIe,                           // supported rates
-                          1, &pAd->PortCfg.SupportedRatesLen,
-                          pAd->PortCfg.SupportedRatesLen, pAd->PortCfg.SupportedRates,
-                          NULL);
+
+		// Build basic frame first
+		MakeOutgoingFrame(OutBuffer,				&FrameLen,
+						  sizeof(MACHDR),			&AssocHdr,
+						  2,						&CapabilityInfo,
+						  2,						&ListenIntv,
+						  1,						&SsidIe,
+						  1,						&pAd->Mlme.SyncAux.SsidLen, 
+						  pAd->Mlme.SyncAux.SsidLen, 	pAd->Mlme.SyncAux.Ssid,
+						  1,						&RateIe,
+						  1,						&pAd->PortCfg.SupRateLen,
+						  pAd->PortCfg.SupRateLen,  pAd->PortCfg.SupRate,
+						  END_OF_ARGS);
+		if (pAd->PortCfg.ExtRateLen != 0)
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, &tmp,
+						1,							&ExtRateIe,
+						1,							&pAd->PortCfg.ExtRateLen,
+						pAd->PortCfg.ExtRateLen,	pAd->PortCfg.ExtRate,							
+						END_OF_ARGS);
+			FrameLen += tmp;
+		}
+		
+		if ((pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPA) && (pAd->PortCfg.WepStatus == Ndis802_11Encryption2Enabled))
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, &tmp,
+						1,						  	&WpaIe,
+						1,						  	&CipherSuiteWpaTkipLen,
+						CipherSuiteWpaTkipLen,	  	&CipherSuiteWpaTkip[0],
+						END_OF_ARGS);
+			FrameLen += tmp;
+			
+			// Add by James 03/06/27
+			// Third add RSN
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &WpaIe, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &CipherSuiteWpaTkipLen, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, CipherSuiteWpaTkip, CipherSuiteWpaTkipLen);
+			VarIesOffset += CipherSuiteWpaTkipLen;
+
+			// Set Variable IEs Length
+			pAd->PortCfg.ReqVarIELen = VarIesOffset;
+			pAd->PortCfg.AssocInfo.RequestIELength = VarIesOffset;
+
+			// OffsetResponseIEs follow ReqVarIE
+			pAd->PortCfg.AssocInfo.OffsetResponseIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION) + pAd->PortCfg.ReqVarIELen;
+			// End Add by James 
+		}
+		
+		else if ((pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPA) && (pAd->PortCfg.WepStatus == Ndis802_11Encryption3Enabled))
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, &tmp,
+						1,						  	&WpaIe,
+						1,						  	&CipherSuiteWpaAesLen,
+						CipherSuiteWpaAesLen,	  	&CipherSuiteWpaAes[0],
+						END_OF_ARGS);
+			FrameLen += tmp;
+			
+			// Add by James 03/06/27
+			// Third add RSN
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &WpaIe, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &CipherSuiteWpaAesLen, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, CipherSuiteWpaAes, CipherSuiteWpaAesLen);
+			VarIesOffset += CipherSuiteWpaAesLen;
+
+			// Set Variable IEs Length
+			pAd->PortCfg.ReqVarIELen = VarIesOffset;
+			pAd->PortCfg.AssocInfo.RequestIELength = VarIesOffset;
+
+			// OffsetResponseIEs follow ReqVarIE
+			pAd->PortCfg.AssocInfo.OffsetResponseIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION) + pAd->PortCfg.ReqVarIELen;
+			// End Add by James 
+		}
+		else if ((pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPAPSK) && (pAd->PortCfg.WepStatus == Ndis802_11Encryption2Enabled))
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, &tmp,
+						1,						  	&WpaIe,
+						1,						  	&CipherSuiteWpaPskTkipLen,
+						CipherSuiteWpaPskTkipLen, 	&CipherSuiteWpaPskTkip[0],
+						END_OF_ARGS);
+			FrameLen += tmp;
+
+			// Add by James 03/06/27
+			// Third add RSN
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &WpaIe, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &CipherSuiteWpaPskTkipLen, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, CipherSuiteWpaPskTkip, CipherSuiteWpaPskTkipLen);
+			VarIesOffset += CipherSuiteWpaPskTkipLen;
+
+			// Set Variable IEs Length
+			pAd->PortCfg.ReqVarIELen = VarIesOffset;
+			pAd->PortCfg.AssocInfo.RequestIELength = VarIesOffset;
+
+			// OffsetResponseIEs follow ReqVarIE
+			pAd->PortCfg.AssocInfo.OffsetResponseIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION) + pAd->PortCfg.ReqVarIELen;
+			// End Add by James 
+		}
+		else if ((pAd->PortCfg.AuthMode == Ndis802_11AuthModeWPAPSK) && (pAd->PortCfg.WepStatus == Ndis802_11Encryption3Enabled))
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, 	&tmp,
+						1,							  	&WpaIe,
+						1,							  	&CipherSuiteWpaPskAesLen,
+						CipherSuiteWpaPskAesLen,	  	&CipherSuiteWpaPskAes[0],
+						END_OF_ARGS);
+			FrameLen += tmp;
+
+			// Add by James 03/06/27
+			// Third add RSN
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &WpaIe, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, &CipherSuiteWpaPskAesLen, 1);
+			VarIesOffset += 1;
+			memcpy(pAd->PortCfg.ReqVarIEs + VarIesOffset, CipherSuiteWpaPskAes, CipherSuiteWpaPskAesLen);
+			VarIesOffset += CipherSuiteWpaPskAesLen;
+
+			// Set Variable IEs Length
+			pAd->PortCfg.ReqVarIELen = VarIesOffset;
+			pAd->PortCfg.AssocInfo.RequestIELength = VarIesOffset;
+
+			// OffsetResponseIEs follow ReqVarIE
+			pAd->PortCfg.AssocInfo.OffsetResponseIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION) + pAd->PortCfg.ReqVarIELen;
+			// End Add by James 
+		}
+		else
+		{
+			// Add by James 03/06/27
+			// Set Variable IEs Length
+			pAd->PortCfg.ReqVarIELen = VarIesOffset;
+			pAd->PortCfg.AssocInfo.RequestIELength = VarIesOffset;
+
+			// OffsetResponseIEs follow ReqVarIE
+			pAd->PortCfg.AssocInfo.OffsetResponseIEs = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION) + pAd->PortCfg.ReqVarIELen;
+			// End Add by James 
+		}
         MiniportMMRequest(pAd, OutBuffer, FrameLen);
             
-        pAd->Mlme.AssocAux.AssocTimer.expires = jiffies + (Timeout * HZ)/1000;
-        add_timer(&pAd->Mlme.AssocAux.AssocTimer);	// in mSec
+		RTMPSetTimer(pAd, &pAd->Mlme.AssocAux.AssocTimer, Timeout);
         pAd->Mlme.AssocMachine.CurrState = ASSOC_WAIT_RSP;
     } 
     else 
@@ -241,6 +442,7 @@ VOID MlmeAssocReqAction(
         pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
         MlmeCntlConfirm(pAd, MT2_ASSOC_CONF, MLME_INVALID_FORMAT);
     }
+
 }
 
 /*
@@ -263,19 +465,28 @@ VOID MlmeReassocReqAction(
 {
     MACADDR              ApAddr;
     MACHDR               ReassocHdr;
-    UCHAR                SsidIe = IE_SSID, RateIe = IE_SUPP_RATES;
+    UCHAR                SsidIe = IE_SSID, RateIe = IE_SUPP_RATES, ExtRateIe = IE_EXT_SUPP_RATES;
     USHORT               CapabilityInfo, ListenIntv;
     ULONG                Timeout;
     ULONG                FrameLen = 0;
+    NDIS_STATUS          NStatus;
+	ULONG			 	 tmp;
     UCHAR               *OutBuffer = NULL;
 
+	// Block all authentication request durning WPA block period
+	if (pAd->PortCfg.bBlockAssoc == TRUE)
+	{
+        DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Block ReAssoc request durning WPA block period!\n");
+        pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
+        MlmeCntlConfirm(pAd, MT2_ASSOC_CONF, MLME_STATE_MACHINE_REJECT);
+	}	
     // the parameters are the same as the association
-    if(MlmeAssocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &ApAddr, &CapabilityInfo, &Timeout, &ListenIntv)) 
+    else if(MlmeAssocReqSanity(pAd, Elem->Msg, Elem->MsgLen, &ApAddr, &CapabilityInfo, &Timeout, &ListenIntv)) 
     {
-        del_timer_sync(&pAd->Mlme.AssocAux.ReassocTimer);
+        RTMPCancelTimer(&pAd->Mlme.AssocAux.ReassocTimer);
 
-        OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_ATOMIC);
-        if(OutBuffer == NULL)
+        NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
+        if(NStatus != NDIS_STATUS_SUCCESS) 
         {
             DBGPRINT(RT_DEBUG_TRACE,"ASSOC - MlmeReassocReqAction() allocate memory failed \n");
             pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
@@ -283,6 +494,8 @@ VOID MlmeReassocReqAction(
             return;
         }
 
+		// Mask out unnecessary capability information
+		CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;  // pAd->PortCfg.SupportedCapabilityInfo;
         pAd->Mlme.AssocAux.CapabilityInfo = CapabilityInfo;
         COPY_MAC_ADDR(&pAd->Mlme.AssocAux.Addr, &ApAddr);
         pAd->Mlme.AssocAux.ListenIntv = ListenIntv;
@@ -291,21 +504,29 @@ VOID MlmeReassocReqAction(
         DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Send RE-ASSOC request...\n");
         MgtMacHeaderInit(pAd, &ReassocHdr, SUBTYPE_REASSOC_REQ, 0, &ApAddr, &ApAddr);
         MakeOutgoingFrame(OutBuffer,            &FrameLen,
-                          sizeof(MACHDR),       (CHAR *)&ReassocHdr,
-                          2,                    &CapabilityInfo,              // capability info
-                          2,                    &ListenIntv,                  // listen interval
-                          MAC_ADDR_LEN,         (char*)&pAd->PortCfg.Bssid,   // current AP address
-                          1,                    &SsidIe,                      // SSID
+                          sizeof(MACHDR),       &ReassocHdr,
+                          2,                    &CapabilityInfo,
+                          2,                    &ListenIntv,
+                          ETH_ALEN,         &pAd->PortCfg.Bssid,
+                          1,                    &SsidIe,
                           1,                    &pAd->PortCfg.SsidLen, 
                           pAd->PortCfg.SsidLen, pAd->PortCfg.Ssid, 
-                          1,                    &RateIe,                      // supported rates
-                          1,                    &pAd->PortCfg.SupportedRatesLen, 
-                          pAd->PortCfg.SupportedRatesLen,   pAd->PortCfg.SupportedRates,
-                          NULL);
+                          1,                    &RateIe,
+						  1,						&pAd->PortCfg.SupRateLen,
+						  pAd->PortCfg.SupRateLen,  pAd->PortCfg.SupRate,
+                          END_OF_ARGS);
+		if (pAd->PortCfg.ExtRateLen != 0)
+		{
+			MakeOutgoingFrame(OutBuffer + FrameLen, &tmp,
+						1,							&ExtRateIe,
+						1,							&pAd->PortCfg.ExtRateLen,
+						pAd->PortCfg.ExtRateLen,	pAd->PortCfg.ExtRate,							
+						END_OF_ARGS);
+			FrameLen += tmp;
+		}
         MiniportMMRequest(pAd, OutBuffer, FrameLen);
             
-        pAd->Mlme.AssocAux.ReassocTimer.expires = jiffies + (Timeout * HZ)/1000;
-        add_timer(&pAd->Mlme.AssocAux.ReassocTimer);	// in mSec
+        RTMPSetTimer(pAd, &pAd->Mlme.AssocAux.ReassocTimer, Timeout); /* in mSec */
         pAd->Mlme.AssocMachine.CurrState = REASSOC_WAIT_RSP;
     } 
     else 
@@ -332,13 +553,14 @@ VOID MlmeDisassocReqAction(
     MACHDR                DisassocHdr;
     CHAR                 *OutBuffer = NULL;
     ULONG                 FrameLen = 0;
-    ULONG                       Timeout = 20;
+    NDIS_STATUS           NStatus;
+    ULONG                 Timeout = 0;
 
     // skip sanity check
     DisassocReq = (MLME_DISASSOC_REQ_STRUCT *)(Elem->Msg);
 
-    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_ATOMIC);
-    if(OutBuffer == NULL)
+    NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
+    if (NStatus != NDIS_STATUS_SUCCESS) 
     {
         DBGPRINT(RT_DEBUG_TRACE, "ASSOC - MlmeDisassocReqAction() allocate memory failed\n");
         pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
@@ -346,20 +568,21 @@ VOID MlmeDisassocReqAction(
         return;
     }
     
-    del_timer_sync(&pAd->Mlme.AssocAux.DisassocTimer);
+    RTMPCancelTimer(&pAd->Mlme.AssocAux.DisassocTimer);
     
     DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Send DISASSOC request\n");
     MgtMacHeaderInit(pAd, &DisassocHdr, SUBTYPE_DISASSOC, 0, &pAd->PortCfg.Bssid, &pAd->PortCfg.Bssid);
-    MakeOutgoingFrame(OutBuffer, &FrameLen, sizeof(MACHDR), (CHAR *)&DisassocHdr, 2, (char*)&DisassocReq->Reason, NULL);
+    MakeOutgoingFrame(OutBuffer,            &FrameLen, 
+                      sizeof(MACHDR),       &DisassocHdr, 
+                      2,                    &DisassocReq->Reason, 
+                      END_OF_ARGS);
     MiniportMMRequest(pAd, OutBuffer, FrameLen);
-    memset(&(pAd->PortCfg.Bssid), 0, MAC_ADDR_LEN);
+    memset(&(pAd->PortCfg.Bssid), 0, ETH_ALEN);
     
     pAd->PortCfg.DisassocReason = REASON_DISASSOC_STA_LEAVING;
     COPY_MAC_ADDR(&pAd->PortCfg.DisassocSta, &DisassocReq->Addr);
-    //StaState(pAd, &DisassocReq->Addr, ASSOC_MODE, ASSOC_MODE_DISASSOC);
 
-    pAd->Mlme.AssocAux.DisassocTimer.expires = jiffies + (Timeout * HZ)/1000;
-    add_timer(&pAd->Mlme.AssocAux.DisassocTimer);	// in mSec
+    RTMPSetTimer(pAd, &pAd->Mlme.AssocAux.DisassocTimer, Timeout); /* in mSec */
     pAd->Mlme.AssocMachine.CurrState = DISASSOC_WAIT_RSP;
 }
 
@@ -378,18 +601,21 @@ VOID PeerAssocRspAction(
     USHORT        CapabilityInfo, Status, Aid;
     UCHAR         Rates[MAX_LEN_OF_SUPPORTED_RATES], RatesLen;
     MACADDR       Addr2;
+    BOOLEAN       ExtendedRateIeExist;
 
-    if (PeerAssocParmSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &Status, &Aid, Rates, &RatesLen)) 
+    if (PeerAssocRspSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &Status, &Aid, Rates, &RatesLen, &ExtendedRateIeExist)) 
     {
         // The frame is for me ?
         if(MAC_ADDR_EQUAL(&Addr2, &pAd->Mlme.AssocAux.Addr)) 
         {
             DBGPRINT(RT_DEBUG_TRACE, "ASSOC - receive ASSOC_RSP to me (status=%d)\n", Status);
-            del_timer_sync(&pAd->Mlme.AssocAux.AssocTimer);
+            RTMPCancelTimer(&pAd->Mlme.AssocAux.AssocTimer);
             if(Status == MLME_SUCCESS) 
             {
                 // go to procedure listed on page 376
-                AssocPostProc(pAd, &Addr2, CapabilityInfo, Aid, Rates, RatesLen);
+				// Mask out unnecessary capability information
+				CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;  // pAd->PortCfg.SupportedCapabilityInfo;
+                AssocPostProc(pAd, &Addr2, CapabilityInfo, Aid, Rates, RatesLen, ExtendedRateIeExist);
             } 
             pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
             MlmeCntlConfirm(pAd, MT2_ASSOC_CONF, Status);
@@ -419,21 +645,21 @@ VOID PeerReassocRspAction(
     UCHAR       Rates[MAX_LEN_OF_SUPPORTED_RATES];
     UCHAR       RatesLen;
     MACADDR     Addr2;
+    BOOLEAN     ExtendedRateIeExist;
 
-    if(PeerAssocParmSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &Status, &Aid, Rates, &RatesLen)) 
+    if(PeerAssocRspSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &CapabilityInfo, &Status, &Aid, Rates, &RatesLen, &ExtendedRateIeExist)) 
     {
         if(MAC_ADDR_EQUAL(&Addr2, &pAd->Mlme.AssocAux.Addr)) // The frame is for me ?
         {
             DBGPRINT(RT_DEBUG_TRACE, "ASSOC - receive REASSOC_RSP to me (status=%d)\n", Status);
-            del_timer_sync(&pAd->Mlme.AssocAux.ReassocTimer);
-            
-            // delete the original association in mac table
-            // StaState(pAd, &pAd->PortCfg.Bssid, ASSOC_MODE, ASSOC_MODE_DISASSOC);
+            RTMPCancelTimer(&pAd->Mlme.AssocAux.ReassocTimer);
             
             if(Status == MLME_SUCCESS) 
             {
+				// Mask out unnecessary capability information
+				CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;  // pAd->PortCfg.SupportedCapabilityInfo;
                 // go to procedure listed on page 376
-                AssocPostProc(pAd, &Addr2, CapabilityInfo, Aid, Rates, RatesLen);
+                AssocPostProc(pAd, &Addr2, CapabilityInfo, Aid, Rates, RatesLen, ExtendedRateIeExist);
             } 
 
             pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
@@ -459,10 +685,12 @@ VOID AssocPostProc(
     IN USHORT CapabilityInfo, 
     IN USHORT Aid, 
     IN UCHAR Rates[], 
-    IN UCHAR RatesLen) 
+    IN UCHAR RatesLen,
+    IN BOOLEAN ExtendedRateIeExist) 
 {
-    // Update MAC status table
-    // StaState(pAd, Addr2, ASSOC_MODE, ASSOC_MODE_ASSOC);
+	ULONG Idx;
+    UCHAR RateIe = IE_SUPP_RATES;
+	UCHAR VarIesOffset;
 
     // 2003/12/11 -  skip the following because experiment show that we can not 
     // trust the "privacy" bit in AssocRsp. We can only trust "Privacy" bit specified in
@@ -477,7 +705,43 @@ VOID AssocPostProc(
 
     // set listen interval
     pAd->PortCfg.DefaultListenCount = pAd->Mlme.AssocAux.ListenIntv;
-    pAd->PortCfg.CurrListenCount = pAd->Mlme.AssocAux.ListenIntv;
+//  pAd->PortCfg.CurrListenCount = pAd->Mlme.AssocAux.ListenIntv;
+
+	// Set New WPA information
+	Idx = BssTableSearch(&pAd->PortCfg.BssTab, Addr2);
+	if (Idx == BSS_NOT_FOUND) 
+	{
+		DBGPRINT(RT_DEBUG_ERROR, "ASSOC - Can't find BSS after receiving Assoc response\n");
+	}
+	else
+	{
+		// Mod by James to fix OID_802_11_ASSOCIATION_INFORMATION
+		pAd->PortCfg.AssocInfo.Length = sizeof(NDIS_802_11_ASSOCIATION_INFORMATION); //+ sizeof(NDIS_802_11_FIXED_IEs); 	// Filled in assoc request
+		pAd->PortCfg.AssocInfo.AvailableResponseFixedIEs =
+			NDIS_802_11_AI_RESFI_CAPABILITIES | NDIS_802_11_AI_RESFI_STATUSCODE | NDIS_802_11_AI_RESFI_ASSOCIATIONID;
+		pAd->PortCfg.AssocInfo.ResponseFixedIEs.Capabilities  = CapabilityInfo;
+		pAd->PortCfg.AssocInfo.ResponseFixedIEs.StatusCode    = MLME_SUCCESS;		// Should be success, add failed later
+		pAd->PortCfg.AssocInfo.ResponseFixedIEs.AssociationId = Aid;
+
+		// Copy BSS VarIEs to PortCfg associnfo structure.
+		// First add Supported rates
+		VarIesOffset = 0;
+		memcpy(pAd->PortCfg.ResVarIEs + VarIesOffset, &RateIe, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ResVarIEs + VarIesOffset, &RatesLen, 1);
+		VarIesOffset += 1;
+		memcpy(pAd->PortCfg.ResVarIEs + VarIesOffset, Rates, RatesLen);
+		VarIesOffset += RatesLen;
+
+		// Second add RSN
+		memcpy(pAd->PortCfg.ResVarIEs + VarIesOffset, pAd->PortCfg.BssTab.BssEntry[Idx].VarIEs, pAd->PortCfg.BssTab.BssEntry[Idx].VarIELen);
+		VarIesOffset += pAd->PortCfg.BssTab.BssEntry[Idx].VarIELen;
+		
+		// Set Variable IEs Length
+		pAd->PortCfg.ResVarIELen = VarIesOffset;
+		pAd->PortCfg.AssocInfo.ResponseIELength = VarIesOffset;
+		// End Mod by James
+	}
 }
 
 /*
@@ -498,28 +762,19 @@ VOID PeerDisassocAction(
     if(PeerDisassocSanity(pAd, Elem->Msg, Elem->MsgLen, &Addr2, &Reason)) 
     {
         if (INFRA_ON(pAd) && MAC_ADDR_EQUAL(&pAd->PortCfg.Bssid, &Addr2)) 
-        {
+        {	
             LinkDown(pAd);
-            // StaState(pAd, &Addr2, ASSOC_MODE, ASSOC_MODE_DISASSOC);
             pAd->Mlme.AssocMachine.CurrState = ASSOC_IDLE;
 
-            // 2002/11/21 -
-            // patch RT2430/RT2420 hangup issue. We suspect this AP DIS-ASSOCIATE frame
-            // is caused by PHY hangup, so we reset PHY, then auto recover the connection.
-            // if this attempt fails, then remains in LinkDown and leaves the problem
-            // to MlmePeriodicExec()
-            // NICPatchRT2430Bug(pAd);
             pAd->RalinkCounters.BeenDisassociatedCount ++;
-			// Remove auto recover effort when disassociate by AP, re-enable for patch 2430 only
             DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Disassociated by AP, Auto Recovery attempt #%d\n", pAd->RalinkCounters.BeenDisassociatedCount);
-            MlmeAutoRecoverNetwork(pAd);
+            MlmeAutoReconnectLastSSID(pAd);
         }
     }
     else
     {
         DBGPRINT(RT_DEBUG_TRACE, "ASSOC - PeerDisassocAction() sanity check fail\n");
     }
-
 }
 
 /*
@@ -616,15 +871,19 @@ VOID Cls3errAction(
     MACHDR                DisassocHdr;
     CHAR                 *OutBuffer = NULL;
     ULONG                 FrameLen = 0;
+    NDIS_STATUS           NStatus;
     USHORT                Reason = REASON_CLS3ERR;
 
-    OutBuffer = kmalloc(MAX_LEN_OF_MLME_BUFFER, GFP_ATOMIC);
-    if(OutBuffer == NULL)
-    	return;
+    NStatus = MlmeAllocateMemory(pAd, (PVOID)&OutBuffer);  //Get an unused nonpaged memory
+    if (NStatus != NDIS_STATUS_SUCCESS) 
+        return;
     
     DBGPRINT(RT_DEBUG_TRACE, "ASSOC - Class 3 Error, Send DISASSOC frame\n");
     MgtMacHeaderInit(pAd, &DisassocHdr, SUBTYPE_DISASSOC, 0, pAddr, &pAd->PortCfg.Bssid);
-    MakeOutgoingFrame(OutBuffer, &FrameLen, sizeof(MACHDR), (CHAR *)&DisassocHdr, 2, (char*)&Reason, NULL);
+    MakeOutgoingFrame(OutBuffer,            &FrameLen, 
+                      sizeof(MACHDR),       &DisassocHdr, 
+                      2,                    &Reason, 
+                      END_OF_ARGS);
     MiniportMMRequest(pAd, OutBuffer, FrameLen);
 
     pAd->PortCfg.DisassocReason = REASON_CLS3ERR;
